// Copyright (C) 2021-2023 David Sugar <tychosoft@gmail.com>.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

package main

import (
	"fmt"
	"os"
	"os/signal"
	"runtime"
	"sync"
	"syscall"
	"time"

	"github.com/alexflint/go-arg"
	"github.com/gofiber/fiber/v2"
	"github.com/gofiber/template/html/v2"
	"gopkg.in/ini.v1"

	"storehouse/lib"
)

// Argument parser....
type Args struct {
	Config  string `arg:"--config" help:"server config file"`
	Host    string `arg:"env:STOREHOUSE_HOST,--host" help:"server host address" default:""`
	Port    uint16 `arg:"env:STOREHOUSE_PORT,--port" help:"server port" default:"8080"`
	Prefix  string `arg:"--prefix" help:"server prefix path"`
	Tls     bool   `arg:"--tls" help:"Server tls mode"`
	Verbose int    `arg:"-v" help:"debugging log level"`
}

// Server config
type Config struct {
	Prefix  string `ini:"prefix"`
	Suffix  string `ini:"suffix"`
	Host    string `ini:"host"`
	Port    uint16 `ini:"port"`
	Views   string `ini:"views"`
	Cache   int    `ini:"cache"`
	Age     int    `ini:"age"`
	Keyfile string `ini:"keyfile"`
	Crtfile string `ini:"crtfile"`
	Tls     bool   `ini:"tls"`
}

// Page config
type Page struct {
	Theme   string `ini:"theme"`
	Lang    string
	Project string
	Version string
	Suffix  string
}

var (
	// binding configs
	workingDir = "/var/lib/storehouse"
	appDataDir = "/usr/share/storehouse"
	etcPrefix  = "/etc"
	logPrefix  = "/var/log"
	version    = "unknown"

	// globals
	args   *Args   = &Args{Prefix: workingDir, Config: etcPrefix + "/storehouse.conf"}
	config *Config = nil
	page   *Page   = nil
	lock   sync.RWMutex
)

func (Args) Version() string {
	return "Version: " + version
}

func (Args) Description() string {
	return "storehouse - provides storehouse access daemon"
}

func init() {
	// parse arguments
	for pos, arg := range os.Args {
		switch arg {
		case "--":
			return
		case "-v":
			os.Args[pos] = "--verbose=1"
		case "-vv":
			os.Args[pos] = "--verbose=2"
		case "-vvv":
			os.Args[pos] = "--verbose=3"
		case "-vvvv":
			os.Args[pos] = "--verbose=4"
		case "-vvvvv":
			os.Args[pos] = "--verbose=5"
		}
	}
	arg.MustParse(args)

	// config service
	logPath := logPrefix + "/storehouse.log"
	lib.Logger(args.Verbose, logPath)
	load()
	err := os.Chdir(config.Prefix)
	if err != nil {
		config.Prefix = appDataDir
		err = os.Chdir(config.Prefix)
		if err != nil {
			lib.Fail(1, err)
		}
	}
}

func load() {
	// default config
	new_config := Config{
		Host:    args.Host,
		Port:    args.Port,
		Cache:   -1,
		Views:   "en",
		Age:     0,
		Prefix:  args.Prefix,
		Suffix:  "tar.gz",
		Tls:     args.Tls,
		Keyfile: "./server.key",
		Crtfile: "./server.crt",
	}

	new_page := Page{
		Theme:   "default",
		Lang:    "en",
		Project: "Storehouse",
		Version: version,
		Suffix:  "tar.gz",
	}

	configs, err := ini.LoadSources(ini.LoadOptions{Loose: true, Insensitive: true}, args.Config, workingDir+"/custom.conf")
	if err == nil {
		// map and reset from args if not default
		configs.Section("server").MapTo(&new_config)
		configs.Section("page").MapTo(&new_page)
		new_page.Suffix = new_config.Suffix
		new_page.Lang = new_config.Views

		if args.Port != 8080 {
			new_config.Port = args.Port
		}

		if len(args.Host) > 0 {
			new_config.Host = args.Host
		}

		if args.Prefix != workingDir {
			new_config.Prefix = args.Prefix
		}

		if new_config.Host == "*" {
			new_config.Host = ""
		}
	} else {
		lib.Error(err)
	}
	lock.Lock()
	defer lock.Unlock()
	config = &new_config
	page = &new_page
}

func main() {
	// setup caching rules
	age := config.Age
	cache := time.Duration(-1)
	if config.Cache > -1 {
		cache = time.Duration(config.Cache) * time.Second
	}

	// setup app and routes
	address := fmt.Sprintf("%s:%v", config.Host, config.Port)
	lib.Debug(3, "prefix=", config.Prefix, ", bind=", address)
	views := appDataDir + "/views_" + config.Views
	if flag, _ := lib.IsDir(views); !flag {
		views = appDataDir + "/views"
	}

	engine := html.New(views, ".html")
	engine.Reload(lib.IsDebug())
	app := fiber.New(fiber.Config{
		ServerHeader:          "Storehouse",
		AppName:               "Storehouse v" + version,
		GETOnly:               true,
		DisableStartupMessage: !lib.IsDebug() || args.Verbose < 1,
		StreamRequestBody:     true,
		Views:                 engine,
	})

	app.Static("/assets", appDataDir+"/assets", fiber.Static{MaxAge: 60})
	app.Get("/", viewIndex)

	if flag, _ := lib.IsDir("./tarballs"); flag {
		app.Static("/tarballs", "./tarballs")
	}

	if flag, _ := lib.IsDir("./archive"); flag {
		app.Static("/archive", "./archive")
	}

	if flag, _ := lib.IsDir("./arch"); flag {
		app.Static("/arch", "./arch", fiber.Static{MaxAge: age, CacheDuration: cache})
	}

	if flag, _ := lib.IsDir("./alpine"); flag {
		app.Static("/alpine", "./alpine", fiber.Static{MaxAge: age, CacheDuration: cache})
	}

	if flag, _ := lib.IsDir("./binary"); flag {
		app.Static("/binary", "./binary", fiber.Static{MaxAge: age, CacheDuration: cache})
	}

	if flag, _ := lib.IsDir("./data"); flag {
		app.Static("/data", "./data")
	}

	if flag, _ := lib.IsDir("./docs"); flag {
		app.Static("/docs", "./docs", fiber.Static{Browse: true})
	}

	if flag, _ := lib.IsDir("./images"); flag {
		app.Static("/images", "./images")
	}

	if flag, _ := lib.IsDir("./screenshots"); flag {
		app.Static("/screenshots", "./screenshots")
	}

	if flag, _ := lib.IsDir("./source"); flag {
		app.Static("/source", "./source")
	}

	// signal handler...
	signals := make(chan os.Signal, 1)
	signal.Notify(signals, os.Interrupt, syscall.SIGTERM, syscall.SIGHUP)
	go func() {
		lib.DaemonLive("start service on ", address)
		defer app.Shutdown()
		defer lib.DaemonStop("stop service")
		for {
			switch <-signals {
			case os.Interrupt: // sigint/ctrl-c
				fmt.Println()
				return
			case syscall.SIGTERM: // normal exit
				return
			case syscall.SIGHUP: // cleanup
				lib.DaemonReload("reload service")
				lib.LoggerRestart()
				runtime.GC()
				load()
				lib.DaemonLive()
			}
		}
	}()

	// start service(s)...
	if config.Tls {
		if err := app.ListenTLS(address, config.Crtfile, config.Keyfile); err != nil {
			lib.Fail(90, err)
		}
	} else if err := app.Listen(address); err != nil {
		lib.Fail(99, err)
	}
}
